from typing import List, Union

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.library.initializers.processors.process_frame import process_frame
from tablate.type.primitives import FrameDivider, ColumnDivider, ContainerBorder, ContainerPadding, ContainerWidth, Background, \
    BackgroundPadding, HtmlPxMultiplier, Multiline, MaxLines, ColumnPadding, TextStyle, TextAlign, TextColor, FrameName, \
    HtmlDefaultColors
from tablate.type.type_input import GridColumnInput, HtmlContainerStylesInput, HtmlGridFrameStylesInput


class Grid(TablateApiItem):
    """
        Create a TablateItem grid frame (single row of multiple columns) to Tablate instance.

        Args:
            columns: Either a list of contents (str, int, float) for columns or a list of column dicts containing contents and styles for columns.
                [required]
            name: Frame name (to prevent duplicate entries).
                [default: None]
            frame_divider: Dividing line at the bottom of the frame.
                [default: 'thick']
            background: Background color for the grig frame.
                [default: None]
            background_padding: Additional padding applied to text if background color set (ASCII only).
                [default: 1]
            multiline: Whether the text may wrap to multiple lines.
                [default: True]
            max_lines: Maximum number of lines to which the text may wrap.
                [default: None]
            column_divider: Dividing line between columns.
                [default: 'thin']
            column_padding: Text padding within each column.
                [default: 1]
            text_style: Text style for grid frame.
                [default: 'normal']
            text_align: Text alignment for grid frame.
                [default: 'left']
            text_color: Text color for grid frame.
                [default: 'None']
            html_px_multiplier: Multiplier applied to HTML px properties.
                [default: 1]
            html_styles: HTML specific styles for grid frame.
                [default: {
                "html_frame_styles": {"html_frame_divider_style": "thick", "html_frame_divider_weight: 1, "html_frame_divider_color": 'black', "html_multiline": True (False for Table frames), "html_max_lines": None, "html_background": None },
                "html_column_styles": {"html_column_divider_style": "thin", "html_column_divider_weight": 1, "html_column_divider_color": "black", "html_padding": "6px"}
                "html_text_styles": {"html_text_style": 'normal', "html_text_align": 'left', "html_text_color": None, "html_text_size": 16 }
                }]
            container_border: outer border style for the Tablate container instance.
                [default: 'thick']
            container_padding: outer padding (in Unicode character width) for the Tablate container instance.
                [default: 1]
            container_width: outer width (in Unicode character width) for the Tablate container instance.
                [default: 120 or terminal width]
            html_container_styles: style dictionary for HTML specific styles for Tablate container instance.
                [default: None]
        """
    def __init__(self,
                 # TablateGrid args
                 columns: List[Union[str, GridColumnInput]],
                 name: FrameName = None,
                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,
                 column_divider: ColumnDivider = None,
                 column_padding: ColumnPadding = None,
                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,
                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlGridFrameStylesInput = None,
                 # TablateApi args
                 html_default_colors: HtmlDefaultColors = None,
                 container_border: ContainerBorder = None,
                 container_padding: ContainerPadding = None,
                 container_width: ContainerWidth = None,
                 html_container_styles: HtmlContainerStylesInput = None) -> None:

        TablateApiBase.__init__(self=self,
                                container_border=container_border,
                                container_padding=container_padding,
                                frame_divider=frame_divider,
                                container_width=container_width,
                                html_container_styles=html_container_styles,
                                html_default_colors=html_default_colors)

        frame_dict = process_frame(frame_name=name,
                                   frame_type="grid",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict



from __future__ import annotations

import sys
from importlib.metadata import version
import importlib

def import_module(module_name):
  """Imports a module without causing a circular import error."""
  if module_name in sys.modules:
    return sys.modules[module_name]

  module = importlib.import_module(module_name)
  sys.modules[module_name] = module

  return module


tablate_module = import_module('tablate.api.Tablate')
text_modules = import_module('tablate.api.modules.Text')
grid_module = import_module('tablate.api.modules.Grid')
table_module = import_module('tablate.api.modules.Table')

concat_module = import_module('tablate.api.functions.concat')

Tablate = tablate_module.Tablate
Text = text_modules.Text
Grid = grid_module.Grid
Table = table_module.Table
concat = concat_module.concat

from tablate.type.type_input import HtmlContainerStylesInput, FrameStylesInput, HtmlFrameStylesInput, ColumnStylesInput, \
    TextStylesInput, HtmlColumnStylesInput, HtmlTextStylesInput, RowsStylesInput, HtmlRowsStylesInput, \
    HtmlColumnInput, BaseColumnInput, GridColumnInput, TableColumnInput, TableRowsInput, BaseStylesInput, \
    TableHeaderFrameStylesInput, TableBodyFrameStylesInput, HtmlStylesInput, HtmlTextFrameStylesInput, \
    HtmlGridFrameStylesInput, HtmlTableFrameStylesInput, HtmlTableHeaderStylesInput

########################################################################################################################
########################################################################################################################
########################################################################################################################


# This will import the modules without causing a circular import error.

try:
    __version__ = version('tablate')
except ModuleNotFoundError:
    pass



# from tablate.type.type_input import HtmlContainerStylesInput, FrameStylesInput, HtmlFrameStylesInput, ColumnStylesInput, \
#     TextStylesInput, HtmlColumnStylesInput, HtmlTextStylesInput, RowsStylesInput, HtmlRowsStylesInput, \
#     HtmlColumnInput, BaseColumnInput, GridColumnInput, TableColumnInput, TableRowsInput, BaseStylesInput, \
#     TableHeaderFrameStylesInput, TableBodyFrameStylesInput, HtmlStylesInput, HtmlTextFrameStylesInput, \
#     HtmlGridStylesInput, HtmlTableStylesInput, HtmlTableHeaderStylesInput, HtmlTableBodyStylesInput


from typing import Union

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.bases.TablateApiItem import TablateApiItem
from tablate.library.initializers.processors.process_frame import process_frame
from tablate.type.primitives import TextAlign, FrameDivider, ContainerBorder, FrameName, TextStyle, TextColor, Background, \
    BackgroundPadding, Multiline, MaxLines, HtmlPxMultiplier, ContainerPadding, ContainerWidth, ColumnPadding, HtmlDefaultColors
from tablate.type.type_input import HtmlTextFrameStylesInput, HtmlContainerStylesInput


class Text(TablateApiItem):
    """
          Create a TablateItem text frame (single row, single column) to Tablate container instance.

          Args:
              text: Text frame content.
                  [required]
              name: Frame name (to prevent duplicate entries).
                  [default: None]
              text_style: Text style for text frame.
                  [default: 'normal']
              text_align: Text alignment for text frame.
                  [default: 'left']
              text_color: Text color for text frame.
                  [default: 'None']
              frame_divider: Dividing line at the bottom of the frame.
                  [default: 'thick']
              frame_padding: Text padding within text frame.
                  [default: 1]
              background: Background color for the text frame.
                  [default: None]
              background_padding: Additional padding applied to text if background color set (ASCII only).
                  [default: 1]
              multiline: Whether the text may wrap to multiple lines.
                  [default: True]
              max_lines: Maximum number of lines to which the text may wrap.
                  [default: None]
              html_px_multiplier: Multiplier applied to HTML px properties.
                  [default: 1]
              html_styles: HTML specific styles for text frame.
                  [default: {
                  "html_frame_styles": {"html_frame_divider_style": "thick", "html_frame_divider_weight: 1, "html_frame_divider_color": 'black', "html_multiline": True (False for Table frames), "html_max_lines": None, "html_background": None },
                  "html_text_styles": {"html_text_style": 'normal', "html_text_align": 'left', "html_text_color": None, "html_text_size": 16 }
                  }]
              container_border: outer border style for the Tablate container instance.
                  [default: 'thick']
              container_padding: outer padding (in Unicode character width) for the Tablate container instance.
                  [default: 1]
              container_width: outer width (in Unicode character width) for the Tablate container instance.
                  [default: 120 or terminal width]
              html_container_styles: style dictionary for HTML specific styles for Tablate container instance.
                  [default: None]
          """
    def __init__(self,
                 # TablateText args
                 text: Union[str, int, float],

                 name: FrameName = None,

                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,
                 frame_padding: ColumnPadding = None,
                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,

                 html_default_colors: HtmlDefaultColors = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlTextFrameStylesInput = None,
                 # TablateApi arge
                 container_border: ContainerBorder = None,
                 container_padding: ContainerPadding = None,
                 container_width: ContainerWidth = None,
                 html_container_styles: HtmlContainerStylesInput = None) -> None:

        TablateApiBase.__init__(self=self,
                                container_border=container_border,
                                container_padding=container_padding,
                                frame_divider=frame_divider,
                                container_width=container_width,
                                html_container_styles=html_container_styles,
                                html_default_colors=html_default_colors)

        frame_dict = process_frame(frame_name=name,
                                   frame_type="text",
                                   frame_args=locals(),
                                   frame_list=self._frame_list,
                                   global_options=self._globals_store.store)

        self._frame_list[frame_dict.name] = frame_dict



from copy import deepcopy
from typing import Union

from tablate.classes.bases.TablateApiBase import TablateApiBase
from tablate.classes.classes import TablateUnion
from tablate.classes.helpers.get_frame import get_frame
from tablate.classes.helpers.list_frame import list_frames
from tablate.library.calcs.gen_frame_name import gen_frame_name


class TablateApiSet(TablateApiBase):

    def list_frames(self):
        """
        List the frames in the Tablate container instance with some details about each frame.
        """
        list_frames(self._frame_list, self._globals_store.store)

    def get_frame(self, selector: Union[int, str], apply_globals: bool = False):
        """
        Gets a frame from the Tablate container instance and returns it as a TablateItem.
        Args:
            selector: The index or name of the frame to select.
            apply_globals: Whether to apply the global defaults and container styles to the returned TablateItem.

        Returns:

        """
        if apply_globals:
            return deepcopy(get_frame(frame_list=self._frame_list, selector=selector, global_options=self._globals_store.args))
        else:
            return deepcopy(get_frame(frame_list=self._frame_list, selector=selector))

    def remove_frame(self, selector: Union[int, str]):
        """
        Deletes a frame.
        Args:
            selector: The index or name of the frame to delete.
        """
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if (type(selector) == int and selector == frame_index) or (type(selector) == str and selector == frame_key):
                del self._frame_list[frame_key]
                break

    def replace_frame(self, selector: Union[int, str], new_frame: TablateUnion, new_name: str = None):
        """
        Replaces a frame with another frame.
        Args:
            selector: The index or name of the frame to replace.
            new_frame: A Tablate Container instance to replace the frame with (if more than one frame in the Tablate container instance, only the first frame will be used).
            new_name: (Optional) The new name of the frame.
        """
        new_frame = deepcopy(new_frame)
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if (type(selector) == int and selector == frame_index) or (type(selector) == str and selector == frame_key):
                new_frame_key, new_frame_item = list(new_frame._frame_list.items())[0]
                new_name = new_name if new_name is not None else new_frame_key
                new_name = gen_frame_name(name=new_name, type=new_frame_item.type, frame_dict=self._frame_list, ensure_unique=True)
                new_frame_item.name = new_name
                new_frame_item.args["name"] = new_name
                self._frame_list = {key if key != frame_key else new_name: value for key, value in self._frame_list.items()}
                self._frame_list[new_name] = new_frame_item

    def rename_frame(self, selector: Union[int, str], new_name: str):
        """
        Renames a frame.
        Args:
            selector: The index or name of the frame to rename.
            new_name: The new name to apply to the frame.
        """
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if (type(selector) == int and selector == frame_index) or (type(selector) == str and selector == frame_key):
                self._frame_list = {key if key != frame_key else new_name: value for key, value in self._frame_list.items()}
                break

    def move_frame(self, from_selector: Union[int, str], to_index: int):
        """
        Move a frame within a Tablate container instance to a new index.
        Args:
            from_selector: The index or name of the frame to move.
            to_index: The index to which the selected frame should be moved.
        """
        selected_frame = get_frame(frame_list=self._frame_list, selector=from_selector, global_options=self._globals_store.args).name[1]
        new_frame_list = {}
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if to_index == frame_index and selected_frame is not None:
                new_frame_list[selected_frame.name] = selected_frame
            if (type(from_selector) == int and from_selector == frame_index) or (type(from_selector) == str and from_selector == frame_key):
                to_index += 1
            else:
                new_frame_list[frame_key] = frame_item
        self._frame_list = new_frame_list

    def insert_frame(self, insert_index: int, new_frame: TablateUnion, new_name: str = None):
        """
        Insert a frame into a Tablate container instance at a specified index.
        Args:
            insert_index: The index at which the new frame should be inserted.
            new_frame: The Tablate container instance to insert.
            new_name: (Optional) The name to apply to the inserted frame.
        """
        new_frame = deepcopy(new_frame)
        new_frame_list = {}
        for frame_index, (frame_key, frame_item) in enumerate(self._frame_list.items()):
            if insert_index == frame_index:
                new_frame_key, new_frame_item = list(new_frame._frame_list.items())[0]
                new_name = new_name if new_name is not None else new_frame_key
                new_name = gen_frame_name(name=new_name, type=new_frame_item.type, frame_dict=self._frame_list, ensure_unique=True)
                new_frame_item.name = new_name
                new_frame_item.args["name"] = new_name
                new_frame_list[new_name] = new_frame_item
            new_frame_list[frame_key] = frame_item
        self._frame_list = new_frame_list





from copy import deepcopy
from typing import List

from tablate.classes.bases.TablateSet import TablateSet
from tablate.classes.classes import TablateUnion
from tablate.type.primitives import FrameDivider, ContainerBorder, ContainerPadding, ContainerWidth, Background, BackgroundPadding, \
    HtmlPxMultiplier
from tablate.type.type_input import HtmlContainerStylesInput, ColumnStylesInput, TextStylesInput, HtmlFrameStylesInput, \
    HtmlColumnStylesInput, HtmlTextStylesInput
from tablate.type.type_store import FrameDictList


def concat(frame_list: List[TablateUnion],
           container_border: ContainerBorder = None,
           container_padding: ContainerPadding = None,
           container_width: ContainerWidth = None,

           frame_divider: FrameDivider = None,
           background: Background = None,
           background_padding: BackgroundPadding = None,

           html_px_multiplier: HtmlPxMultiplier = None,
           html_container_styles: HtmlContainerStylesInput = None,

           column_styles: ColumnStylesInput = None,
           text_styles: TextStylesInput = None,

           html_frame_styles: HtmlFrameStylesInput = None,

           html_column_styles: HtmlColumnStylesInput = None,
           html_text_styles: HtmlTextStylesInput = None):
    """
    Merge a list of Tablate container instances into a new Tablate container instance. (Useful for IPython environment.)

    Args:
        frame_list: A list of Tablate container instances.
        container_border: outer border style for the Tablate container instance.
            [default: 'thick']
        container_padding: outer padding (in Unicode character width) for the Tablate container instance.
            [default: 1]
        container_width: outer width (in Unicode character width) for the Tablate container instance.
            [default: 120 or terminal width]
        frame_divider: default dividing line between frames within the Tablate container instance.
            [default: 'thick']
        background: default background for frames withing the Tablate container instance.
            [default: None]
        background_padding: additional padding applied to cells with a background color set (ASCII only).
            [default: 1]
        html_px_multiplier: Multiplier applied to HTML px properties.
            [default: 1]
        html_container_styles: style dictionary for HTML specific styles for Tablate container instance.
            [default: None]
        column_styles: style dictionary for default base column styles.
            [default: {"column_divider": 'thin', "padding": 1, "background_padding": 1}]
        text_styles: style dictionary for default base text styles.
            [default: {
            "text_style": 'normal',
            "text_align": 'left',
            "text_color": None
            }]
        html_frame_styles: style dictionary for default HTML specific frame styles.
            [default: {\n
            "html_frame_divider_style": "thick",
            "html_frame_divider_weight: 1,
            "html_frame_divider_color": 'black',
            "html_multiline": True (False for Table frames),
            "html_max_lines": None,
            "html_background": None
            }]
        html_column_styles: style dictionary for default HTML specific column styles.
            [default: {
            "html_column_divider_style": 'thin',
            "html_column_divider_weight": 1,
            "html_column_divider_color": 'black',
            "html_padding": 6px
            }]
        html_text_styles: style dictionary for default HTML specific text styles.
            [default: {
            "html_text_style": 'normal',
            "html_text_align": 'left',
            "html_text_color": None,
            "html_text_size": 16
            }]

    Returns: A new TablateSet instance.

    """
    input_dict_list: List[FrameDictList] = []

    for frame_set in frame_list:
        input_dict_list.append(deepcopy(frame_set._frame_list))

    args = deepcopy(locals())

    args["frame_list"] = input_dict_list
    del args["input_dict_list"]
    del args["frame_set"]

    return TablateSet(**args)

# todo: this needs to have process frame applied (probably in TabletSet)
